#ifndef HCAL_AMC13_AMC13_HH_INCLUDED
#define HCAL_AMC13_AMC13_HH_INCLUDED 1

#include <string>
#include <vector>
#include <map>
#include <stdint.h>
#include <stdlib.h>
#include <dirent.h>
#include <assert.h>

#include "hcal/amc13/ipDev.hh"
#include "hcal/amc13/AMC13_env.hh"
#include "hcal/amc13/AMC13_verify.hh"

#define MAX_WORDS_PER_BLOCK_READ 256

namespace cms {

  typedef uint32_t rv; 

  class AMC13 {
  public:
    //Enumerations used by the AMC13 class
    enum { T2 = 0, T1 = 1 };
    enum { spartan = T2, virtex = T1, kintex = T1 };
    enum { virtex_t1 = 0, kintex_t1 = 1 };

    //Constructors
    AMC13(const std::string&, const std::string&, int&,
	  const std::string&, const std::string&, int&,
	  bool controlHub = true); 
    AMC13(const std::string&, const std::string&,
	  const std::string&, const std::string&,
	  bool controlHub = true);
    AMC13(AMC13_env*);
    AMC13(ipDev*, ipDev*);
    AMC13() { };

    //Destructor
    ~AMC13();

    // General Read
    rv read(int, const std::string&) throw(ipDev::exception);
    rv readAddress(int, uint32_t) throw(ipDev::exception);
    std::vector<rv> readBlockAddress(int, uint32_t, size_t) throw(ipDev::exception);
    size_t readBlockAddress(int, uint32_t, size_t, uint32_t*, size_t) throw(ipDev::exception);
    std::vector<rv> readFifoAddress(int, uint32_t, size_t) throw(ipDev::exception);
    std::map<std::string, rv> readAllReadable(int) throw(ipDev::exception);

    // General Write
    void write(int, const std::string&, uint32_t) throw(ipDev::exception);
    void writeAddress(int, uint32_t, uint32_t) throw(ipDev::exception);
    void writeBlockAddress(int, uint32_t, std::vector<uint32_t>) throw(ipDev::exception);
    void writeFifoAddress(int, uint32_t, std::vector<uint32_t>) throw(ipDev::exception);
    void writeQueueAddress(int, uint32_t, std::vector<uint32_t>, size_t) throw(ipDev::exception);
    void writeTest(int) throw(ipDev::exception);

    // General
    void reset(int) throw(ipDev::exception);

    //Initialization
    void AMCInputEnable(std::string, bool slotbased=false) throw(ipDev::exception);
    void enableAllTTC() throw(ipDev::exception);
    void daqLinkEnable(bool) throw(ipDev::exception);
    void fakeDataEnable(bool) throw(ipDev::exception);
    void localTtcSignalEnable(bool) throw(ipDev::exception);
    void genInternalL1AsEnable(bool) throw(ipDev::exception);
    void ttcRxEnable(bool) throw(ipDev::exception);
    void monBufBackPressEnable(bool) throw(ipDev::exception);

    //Control
    void sendLocalEvnOrnReset(uint32_t a, uint32_t b) throw(ipDev::exception);
    void genInternalPeriodicL1As(bool b) throw(ipDev::exception);
    void enableDaqLinkSenderReceiver() throw(ipDev::exception);
    void saveReceivedDaqData(bool b) throw(ipDev::exception);
    void megaMonitorScale(bool b) throw(ipDev::exception);
    void setPreScaleFactor(uint32_t noZos) throw(ipDev::exception);
    void genInternalSingleL1A(uint32_t n) throw(ipDev::exception);
    void setTrigType(uint32_t type) throw(ipDev::exception);
    void setLocalL1APeriod(uint32_t n) throw(ipDev::exception);
    void setFEDid(uint32_t id) throw(ipDev::exception);
    void startRun() throw(ipDev::exception);
    void endRun() throw(ipDev::exception);

    // DAQ
    rv nextEventSize() throw(ipDev::exception); 
    rv readNextEvent(uint32_t*, uint32_t) throw(ipDev::exception); 
    rv readNextEventNoAdvance(uint32_t*, uint32_t) throw(ipDev::exception);

    //Accessor functions
    ipDev* fpga(int chip) { return fpga_.at(chip); }
    std::string getT2ID()  const { return T2_ID;   };
    std::string getT1ID()  const { return T1_ID;   };
    std::string getT2URI() const { return T2_URI;  };
    std::string getT1URI() const { return T1_URI;  };
    std::string getT2ADD() const { return T2_ADD;  };
    std::string getT1ADD() const { return T1_ADD;  };
    int getKinVir() const { return kinOrVir; };

  private:
    //Functions used by constructors/destructors to build/kill ipDev objects
    void  kintexOrVirtex();
    void  kintexOrVirtex(const std::string&, const std::string&, 
			const int&, const bool&);
    void buildCorrectT2(const std::string&, const std::string&, 
			int&, const bool&);
    void buildCorrectT1(const std::string&, const std::string&, 
			int&, const bool&);
    void buildCorrectIpDevs(const std::string&, const std::string&, int&,
			    const std::string&, const std::string&, int&, 
			    const bool&);
    void buildT2(const std::string&, const std::string&, 
		 const int&, const bool&);
    void buildT1(const std::string&, const std::string&, 
		 const int&, const bool&);
    void buildIpDevs(const std::string&, const std::string&, const int&,
		     const std::string&, const std::string&, const int&,
		     const bool&);
    void assignVars();
    void killT2();
    void killT1();
    void killIpDevs();

    //Function to help with input enable
    rv parseInputEnableList(std::string, bool);

    //Private class objects
    std::vector<ipDev*> fpga_;

    //Private variables
    rv ret;
    std::vector<rv> retVec;
    std::map<std::string, rv> retmap;

    //Private identification variables
    std::string T2_ID,  T1_ID;
    std::string T2_URI, T1_URI;
    std::string T2_ADD, T1_ADD;
    int kinOrVir;

    //For handling addresses
    uint32_t addr;

  };

}

#endif //HCAL_AMC13_AMC13_HH_INCLUDED
